import {
  geoNaturalEarth1,
  geoPath,
  geoGraticule,
  contourDensity,
  scaleSequential,
  interpolateCubehelixDefault,
  interpolateReds,
  max,
} from 'd3';

const oceanColor = '#001338';
const internalBordersColor = '#003194';
const internalBordersStrokeWidth = 0.5;

const populationThresholdForLabels = 7000000;
// const populationThresholdForLabels = 1000000;

const projection = geoNaturalEarth1();
const path = geoPath(projection);
const graticule = geoGraticule();

export const map = (
  selection,
  {
    countries,
    land,
    countryBorders,
    cities,
    width,
    height,
  },
) => {
  // Project Cities
  for (const d of cities) {
    const [x, y] = projection([d.lng, d.lat]);
    d.x = x;
    d.y = y;
  }

  // Cities to label
  const filteredCities = cities.filter(
    (d) => d.population > populationThresholdForLabels,
  );

  // Compute Contours
  const contours = contourDensity()
    .x((d) => d.x)
    .y((d) => d.y)
    .weight((d) => d.population)
    .size([width, height])
    .thresholds(71)
    .cellSize(1)
    // The top number is tweakable with Alt+Drag in VizHub.
    .bandwidth(392 / 100)(cities);

  // Colors
  const colorScale = scaleSequential((t) => {
    const t1 =
      (t *
        // Controls "exposure"
        701) /
        100 +
      // Controls the luminance of
      // the first contour line
      2 / 100;
    // console.log(t1);

    return t1 < 1
      ? // Start with CubeHelix as the first level
        interpolateCubehelixDefault(t1)
      : // Loop around reds as subsequent levels
        interpolateReds(t1 % 1);
  })
    .domain([0, max(contours, (d) => d.value)])
    .clamp(true);

  // Land Outline Clipping Path
  // See https://developer.mozilla.org/en-US/docs/Web/SVG/Tutorial/Clipping_and_masking
  selection
    .selectAll('defs')
    .data([null])
    .join('defs')
    .selectAll('clipPath')
    .data([null])
    .join('clipPath')
    .attr('id', 'land-clipping-path')
    .selectAll('path.land')
    .data(land.features)
    .join('path')
    .attr('d', path)
    .attr('class', 'land')
    .attr('fill', '#000000');

  // Globe outline
  selection
    .selectAll('path.outline')
    .data([null])
    .join('path')
    .attr('class', 'outline')
    .attr('d', path(graticule.outline()))
    .attr('fill', oceanColor);

  // Graticules
  selection
    .selectAll('path.graticule')
    .data([null])
    .join('path')
    .attr('class', 'graticule')
    .attr('d', path(graticule()))
    .attr('fill', 'none')
    .attr('stroke', '#000000')
    .attr('stroke-width', 0.5);

  // Background within land
  selection
    .selectAll('rect')
    .data([null])
    .join('rect')
    .attr('x', '0')
    .attr('y', '0')
    .attr('width', width)
    .attr('height', height)
    .attr('fill', '#000000')
    .attr('clip-path', 'url(#land-clipping-path)');

  // Contours
  selection
    .selectAll('path.contour')
    .data(contours)
    .join('path')
    .attr('d', geoPath())
    .attr('class', 'contour')
    .attr('stroke', 'none')
    .attr('fill', (d) => colorScale(d.value))
    .attr('clip-path', 'url(#land-clipping-path)');
  // TODO use countries mask for contours
  // Countries mask

  // Internal Country Borders
  selection
    .selectAll('path.countries')
    .data([...land.features, countryBorders])
    // .data([countryBorders])
    .join('path')
    .attr('d', path)
    .attr('class', 'countries')
    .attr('fill', 'none')
    .attr('stroke', internalBordersColor)
    .attr('stroke-width', internalBordersStrokeWidth)
    .attr('stroke-linejoin', 'round');

  // selection
  //   .selectAll('circle')
  //   .data(filteredCities)
  //   .join('circle')
  //   .attr('cx', (d) => d.x)
  //   .attr('cy', (d) => d.y)
  //   .attr('r', 1);

  selection
    .selectAll('g.labels')
    .data([null])
    .join('g')
    .attr('class', 'labels')
    .attr('text-anchor', 'middle')
    .attr('fill', '#FFFFFF')
    .attr('stroke', '#000000')
    .attr('paint-order', 'stroke')
    .attr('font-family', 'sans-serif')
    .attr('stroke-linejoin', 'round')
    .selectAll('text')
    .data(filteredCities)
    .join('text')
    .attr('alignment-baseline', 'middle')
    .attr('x', (d) => d.x)
    .attr('y', (d) => d.y)
    .attr('font-size', (d) =>
      Math.sqrt(d.population / 800000),
    )
    .attr('stroke-width', (d) => d.population / 30000000)
    .text((d) => d.city);

  // TODO Filter out overlapping labels by:
  // - indexing the labels in an R-Tree
  // - sorting the cities by population
  // - for each label
  const boundingBoxes = selection
    .selectAll('text')
    .nodes()
    .map((node) => node.getBBox());
};
